/*
 * PointList.java
 *
 * Created on August 2, 2006, 3:31 PM
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package sddm.data;

import sddm.*;
import java.util.*;
import java.io.*;

/**
 * The class represents an ordered set of Points.  For an unordered set with no repeated elements, use {@link PointSet}.
 * @author Erik J. Tollerud
 */
public class PointList implements Iterable<float[]>,Cloneable {
    
    /**
     * The list of 4-element arrays with the points.
     */
    protected java.util.ArrayList<float[]> list;
    //format is {xmin,xmax,xavg,ymin,ymax,yavg,zmin,zmax,zavg,wmin,wmax,wavg}
    float[] stats;
    
    /** Creates a new instance of PointList that is empty*/
    public PointList() {
        list = new java.util.ArrayList<float[]>();
        
    }
    
    /**
     * Creates a new instance of PointList from an array of float arrays.
     * @param points the points to use in generating this PointList.  Must all be 3-element or 4-element.
     */
    public PointList(float[][] points) {
        list = new java.util.ArrayList<float[]>();
        this.setPoints(points);
    }
    
    /**
     * Creates a new instance of PointList from a file (see loadPointsFromFile(File).
     * @param pointFile the file from which to load the points.
     * @throws sddm.FileFormatException If the file does not have numbers, or the number of numbers isn't divisible by 3.
     * @throws java.io.FileNotFoundException If the specified file doesn't exist.
     */
    public PointList(File pointFile) throws FileFormatException, FileNotFoundException {
        list = new java.util.ArrayList<float[]>();
        this.loadPointsFromFile(pointFile);
    }
    
    /**
     * Creates a new instance of PointList from a file (see loadPointsFromFile(File, boolean).
     * @param pointFile the file from which to load the points.
     * @param threeElement if true, only sets x,y, and z cooordinates from the file (w is set to 1.0).
     * @throws sddm.FileFormatException If the file does not have numbers, or the number of numbers isn't divisible by 3 or 4.
     * @throws java.io.FileNotFoundException If the specified file doesn't exist.
     */
    public PointList(File pointFile, boolean threeElement) throws FileFormatException, FileNotFoundException {
        list = new java.util.ArrayList<float[]>();
        this.loadPointsFromFile(pointFile, threeElement);
    }
    
    /**
     * Creates a new instance of PointList from a PointSet. Only shallowly copies elements.
     * @param ps the set to use for the points
     */
    PointList(PointSet ps) {
        this();
        for (float[] f : ps)
            this.addPoint(f);
    }
    
    /**
     * Shallowly copies this List to a set (will eliminate duplicate points).
     * @return a PointSet with the same points as this PointList.
     */
    public PointSet toSet() {
        return new PointSet(this);
    }
        
    /**
     * Retrieves the number of elements.
     * @return the number of points in this PointList
     */
    public int size() {
        return list.size();
    }
    
    /**
     * Determines if this PointList is empty.
     * @return true if there are no points in this list.
     */
    public boolean isEmpty() {
        return list.isEmpty();
    }
    
    /**
     * Remove all points from this PointList
     */
    public synchronized void clear() {
        list.clear();
        stats = null;
    }
    
    /**
     * Sets the point at the speicified index of this PointList
     * @param i the index to replace with the specified point
     * @param point A 3 or 4-element array to copy values in from.
     * @throws sddm.SDDMException if the point array is not 3 or 4 elements long.
     * @throws java.lang.IndexOutOfBoundsException If the speicified index is larger than the size of this PointList.
     */
    public synchronized void setPoint( int i, float[] point) throws SDDMException, IndexOutOfBoundsException {
        if ( point.length == 4 )
            this.list.set(i,point);
        else if (point. length == 3) {
            float[] newPt = new float[4];
            newPt[0] = point[0];
            newPt[1] = point[1];
            newPt[2] = point[2];
            newPt[3] = 1.0f;
            this.list.set(i,newPt);
        }
        else
            throw new SDDMException("Tried to set point in PointList that does not have 3 or 4 elements.");
    }
    
    /**
     * Sets the point at the speicified index of this PointList
     * @param i the index to replace with the specified point
     * @param x the x-coordinate to use at the specified index.
     * @param y the y-coordinate to use at the specified index.
     * @param z the z-coordinate to use at the specified index.
     * @throws java.lang.IndexOutOfBoundsException If the speicified index is larger than the size of this PointList.
     */
    public synchronized void setPoint( int i, float x, float y, float z) throws IndexOutOfBoundsException {
        
        float[] point = new float[4];
        point[0] = x;
        point[1] = y;
        point[2] = z;
        point[3] = 1.0f;
        this.list.set(i, point);

    }
    
    /**
     * Sets the point at the speicified index of this PointList
     * @param i the index to replace with the specified point
     * @param x the x-coordinate to use at the specified index.
     * @param y the y-coordinate to use at the specified index.
     * @param z the z-coordinate to use at the specified index.
     * @param w the w-coordinate to use at the specified index.
     * @throws java.lang.IndexOutOfBoundsException If the specified index is larger than the size of this PointList.
     */
    public synchronized void setPoint( int i, float x, float y, float z, float w) throws IndexOutOfBoundsException {
        
        float[] point = new float[4];
        point[0] = x;
        point[1] = y;
        point[2] = z;
        point[3] = w;
        this.list.set(i, point);

    }
        
    /**
     * Sets the points in this PointList to the specified points, clearing the exisitng points.
     * @param points The points to use as an array of {x,y,z} or {x,y,z,w} array.
     * @throws sddm.SDDMException If one of the points does not have 3 or 4 elements.
     */
    public synchronized void setPoints(float[][] points) throws SDDMException {
        list.clear();
        for (float[] f : points)
            this.addPoint(f);
        stats = null;
    }
    
    /**
     * 
     * Appends the supplied points to this PointList.
     * @param points The points to add as an array of {x,y,z} or {x,y,z,w} array.
     * @throws sddm.SDDMException If one of the points does not have 3 or 4 elements.
     */
    public synchronized void addPoints(float[][] points) throws SDDMException {
        for (float[] f : points)
            this.addPoint(f);
        stats = null;
    }
    
    /**
     * 
     * Loads a list of points from a file that has only numbers in x0 y0 z0 x1 y1 z1 x2 y2 z2 ... format.
     * @param pointFile the file from which to load the points.
     * @throws sddm.FileFormatException If the file does not have numbers, or the number of numbers isn't divisible by 3.
     * @throws java.io.FileNotFoundException If the specified file doesn't exist.
     */
    public void loadPointsFromFile(File pointFile) throws FileFormatException, FileNotFoundException {
        //assume 3D if not specified
        this.loadPointsFromFile(pointFile,true);
    }
    
    /**
     * 
     * Loads a list of points from a file that has only numbers in x0 y0 z0 x1 y1 z1 x2 y2 z2 ...  or x0 y0 z0 w0 x1 y1 z1 w1 x2 y2 z2 w2 ... format.
     * @param pointFile the file from which to load the points.
     * @param threeElement if true, only sets x,y, and z cooordinates from the file (w is set to 1.0).
     * @throws sddm.FileFormatException If the file does not have numbers, or the number of numbers isn't divisible by 3 or 4.
     * @throws java.io.FileNotFoundException If the specified file doesn't exist.
     */
     public void loadPointsFromFile(File pointFile, boolean threeElement) throws FileFormatException, FileNotFoundException {
        
         Scanner s = new Scanner(pointFile);
     
         if (threeElement) {
             while(s.hasNextFloat()) {
                 float x = s.nextFloat();
                 float y = s.nextFloat();
                 if (!s.hasNextFloat())
                     throw new FileFormatException("Number of elements in point list file not divisible by 3.");
                     float z = s.nextFloat();
                 this.addPoint(x,y,z);
             }
         } else {
             while(s.hasNextFloat()) {
                 float x = s.nextFloat();
                 float y = s.nextFloat();
                 float z = s.nextFloat();
                 if (!s.hasNextFloat())
                     throw new FileFormatException("Number of elements in point list file not divisible by 4.");
                     float w = s.nextFloat();
                 this.addPoint(x,y,z,w);
             }
         }
         stats = null;
     }
         
    /**
     * copies in elements
     * @param point the 3D or 4D point in {x,y,z} pr {x,y,z,w} format.
     * @throws sddm.SDDMException if there are not 3 or 4 components in the provided point.
     */
    public synchronized void addPoint(float[] point) throws SDDMException {
      if (point.length == 4) 
          this.addPoint(point[0],point[1],point[2],point[3]);
      else if (point.length == 3) 
          this.addPoint(point[0],point[1],point[2]);
      else
          throw new SDDMException("tried to load point that is not 3 or 4 elements");
      stats = null;
    }
    
    /**
     * 
     * Appends a point with the specified coordinates to this PointList.
     * @param x The x-coordinate of the point to add.
     * @param y The y-coordinate of the point to add.
     * @param z The z-coordinate of the point to add.
     */
    public synchronized void addPoint(float x, float y, float z) {
        float[] element = new float[4];
        element[0] = x;
        element[1] = y;
        element[2] = z;
        element[3] = 1.0f;
        
        list.add(element);
        stats = null;
    }
    
    /**
     * 
     * Appends a point with the specified coordinates to this PointList.
     * @param x The x-coordinate of the point to add.
     * @param y They-coordinate of the point to add.
     * @param z The z-coordinate of the point to add.
     * @param w The w-coordinate of the point to add.
     */
    public synchronized void addPoint(float x, float y, float z, float w) {
        float[] element = new float[4];
        element[0] = x;
        element[1] = y;
        element[2] = z;
        element[3] = w;
        
        list.add(element);
        stats = null;
    }
    
    /**
     * 
     * Removes the point at the specified index.
     * @param index the index from which to retrieve the point.
     * @return the point at the specified index as a 4-element array.
     * @throws java.lang.IndexOutOfBoundsException If there is no point at the specified index.
     */
    public float[] removePoint(int index) throws IndexOutOfBoundsException  {
        return list.remove(index);
    }
    
    /**
     * Looks for the first index at which the x, y, and z coordinates match the requested values.
     * @param x the x-coordinate to be matched.
     * @param y the y-coordinate to be matched.
     * @param z the z-coordinate to be matched.
     * @return the index that matches the specified coordinates, or -1 if there is no such point.
     */
    public int indexOf(float x, float y, float z) {
        int index = -1;
        for (int i = 0; i < this.list.size(); ++i) {
            float[] f = this.list.get(i);
            if (f[0] == x && f[1] == y && f[2] == z)
                index = i;
        }
        return index;
    }
    
    /**
     * 
     * Looks for the first index at which the x, y, z, and w coordinates match the requested values.
     * @param x the x-coordinate to be matched.
     * @param y the y-coordinate to be matched.
     * @param z the z-coordinate to be matched.
     * @param w the w-coordinate to be matched.
     * @return the index that matches the specified coordinates, or -1 if there is no such point.
     */
    public int indexOf(float x, float y, float z, float w) {
        int index = -1;
        for (int i = 0; i < this.list.size(); ++i) {
            float[] f = this.list.get(i);
            if (f[0] == x && f[1] == y && f[2] == z && f[3] == w)
                index = i;
        }
        return index;
    }
    
    /**
     * Looks for the first index at which the coordinates match the requested values.
     * @param point a 3 or 4-element vector to use as lookup values.
     * @return the index of the first point that shares the elements specified in point, or -1 if there is no such element.
     * @throws sddm.SDDMException if the supplied point does not have 3 or 4 elements in the array.
     */
    public int indexOf(float[] point) throws SDDMException {
        if ( point.length ==3 ) {
            int index = -1;
            for (int i = 0; i < this.list.size(); ++i) {
                float[] f = this.list.get(i);
                if (f[0] == point[0] && f[1] == point[1] && f[2] == point[2] )
                    index = i;
            }
            return index;
        }
        else if ( point.length == 4 ) {
            int index = -1;
            for (int i = 0; i < this.list.size(); ++i) {
                float[] f = this.list.get(i);
                if (f[0] == point[0] && f[1] == point[1] && f[2] == point[2] && f[3] == point[3])
                    index = i;
            }
            return index;
        }
        else 
            throw new SDDMException("Tried to find index of array that did not have 3 or 4 elements.");
    }
    
    /**
     * Retrieves a reference to the point in this PointList at the specified index.
     * @param index the index from which to retrieve a point.
     * @return a reference to the point as a 4-element {x,y,z,w} array.
     * @throws java.lang.IndexOutOfBoundsException if there is no point at the specified index.
     */
    public float[] getPoint(int index) throws IndexOutOfBoundsException {
        return list.get(index);
    }
    
    
    
    /**
     * 
     * Retrieves a reference to the last point in this PointList.
     * @return a reference to the last point as a 4-element {x,y,z,w} array.
     */
    public float[] getLastPoint() {
        return list.get(list.size()-1);
    }
    
    /**
     * Copies the point at the specified index and returns the copy.
     * @param index the index from which to retrieve a point.
     * @return an element-wise copy of the element at the specified index as an {x,y,z,w} array.
     */
    public float[] copyPoint(int index) {
        return list.get(index).clone();
    }
    
    /**
     * 
     * Retrieves the x-coordinate of the point at the specified index.
     * @param index the from which index to retrieve the coordinate.
     * @return the x-coordinate.
     * @throws java.lang.IndexOutOfBoundsException If there is no element with the specified index.
     */
    public float getX(int index) throws IndexOutOfBoundsException {
        return list.get(index)[0];
    }
    
    /**
     * 
     * Retrieves the y-coordinate of the point at the specified index.
     * @param index the from which index to retrieve the coordinate.
     * @return the y-coordinate.
     * @throws java.lang.IndexOutOfBoundsException If there is no element with the specified index.
     */
    public float getY(int index) throws IndexOutOfBoundsException {
        return list.get(index)[1];
    }
    
    /**
     * 
     * Retrieves the z-coordinate of the point at the specified index.
     * @param index the from which index to retrieve the coordinate.
     * @return the z-coordinate.
     * @throws java.lang.IndexOutOfBoundsException If there is no element with the specified index.
     */
    public float getZ(int index) throws IndexOutOfBoundsException {
        return list.get(index)[2];
    }
    
    /**
     * 
     * Retrieves the w-coordinate of the point at the specified index.
     * @param index the from which index to retrieve the coordinate.
     * @return the w-coordinate.
     * @throws java.lang.IndexOutOfBoundsException If there is no element with the specified index.
     */
    public float getW(int index) throws IndexOutOfBoundsException {
        return list.get(index)[3];
        
    }
    
    /**
     * 
     * Generates the iterator for this PointList.
     * @return an Iterator that references 4-element point arrays.
     */
    public Iterator<float[]> iterator() {
        return list.iterator();
    }
    
    /**
     * 
     * Generates an iterator that returns the x-coordinates of the points in this PointList.
     * @return an iterator that visits each of the x-coordinates in order.
     */
    public Iterator<Float> xIterator() {
        return new ElemIter(0);
    }
    
    /**
     * 
     * Generates an iterator that returns the y-coordinates of the points in this PointList.
     * @return an iterator that visits each of the y-coordinates in index order.
     */
    public Iterator<Float> yIterator() {
        return new ElemIter(1);
    }
    
    /**
     * 
     * Generates an iterator that returns the z-coordinates of the points in this PointList.
     * @return an iterator that visits each of the z-coordinates in index order.
     */
    public Iterator<Float> zIterator() {
        return new ElemIter(2);
    }
    
    /**
     * 
     * Generates an iterator that returns the w-coordinates of the points in this PointList.
     * @return an iterator that visits each of the yw-coordinates in index order.
     */
    public Iterator<Float> wIterator() {
        return new ElemIter(3);
    }
    
    /**
     * Randomly adjust each points' coordinates by (at most) the specified amount
     * @param scale the largest allowed perturbation
     * @param doW if flase, does not perturb the w-coordinate
     * @param rand the random number generator to use 
     */
    public void randomPerturbation(float scale, boolean doW, Random rand) {
        for (float[] f : this.list) {
            f[0]+=scale*(rand.nextFloat()*2.0f - 1.0f);
            f[1]+=scale*(rand.nextFloat()*2.0f - 1.0f);
            f[2]+=scale*(rand.nextFloat()*2.0f - 1.0f);
            if (doW)
                f[3]+=scale*(rand.nextFloat()*2.0f - 1.0f);
        }
        stats = null;
    }
    
    /**
     * Randomly adjust each points' coordinates by (at most) the specified amount
     * @param scale the largest allowed perturbation
     * @param doW if flase, does not perturb the w-coordinate
     * @param seed the seed to use in generating the default Java RNG
     */
    public void randomPerturbation(float scale, boolean doW, long seed) {
        this.randomPerturbation(scale, doW, new java.util.Random(seed));
    }
    
    /**
     * Randomly adjust each points' coordinates by (at most) the specified amount
     * @param scale the largest allowed perturbation
     * @param doW if flase, does not perturb the w-coordinate
     */
    public void randomPerturbation(float scale, boolean doW) {
        this.randomPerturbation(scale, doW, new java.util.Random());
    }
    
    /**
     * 
     * Retrieves the minimum x value in this PointList.  Will Take longer on the first call after a change has been made.
     * @return the smallest x value in this PointList.
     */
    public float getXMin() {
        if (stats == null)
            makeStats();
        return stats[0];
    }
    
    /**
     * 
     * Retrieves the maximum x value in this PointList.  Will Take longer on the first call after a change has been made.
     * @return the largest x value in this PointList.
     */
    public float getXMax() {
        if (stats == null)
            makeStats();
        return stats[1];
    }
    
    /**
     * 
     * Retrieves the average x value in this PointList.  Will Take longer on the first call after a change has been made.
     * @return The average of the x values in this PointList.
     */
    public float getXAvg() {
        if (stats == null)
            makeStats();
        return stats[2];
    }
    
    /**
     * 
     * 
     * Retrieves the minimum y value in this PointList.  Will Take longer on the first call after a change has been made.
     * @return the smallest y value in this PointList.
     */
    public float getYMin() {
        if (stats == null)
            makeStats();
        return stats[3];
    }
    
    /**
     * 
     * 
     * Retrieves the maximum y value in this PointList.  Will Take longer on the first call after a change has been made.
     * @return the largest y value in this PointList.
     */
    public float getYMax() {
        if (stats == null)
            makeStats();
        return stats[4];
    }
    
    /**
     * 
     * 
     * Retrieves the average y value in this PointList.  Will Take longer on the first call after a change has been made.
     * @return The average of the y values in this PointList.
     */
    public float getYAvg() {
        if (stats == null)
            makeStats();
        return stats[5];
    }
    
    /**
     * 
     * 
     * Retrieves the minimum z value in this PointList.  Will Take longer on the first call after a change has been made.
     * @return the smallest z value in this PointList.
     */
    public float getZMin() {
        if (stats == null)
            makeStats();
        return stats[6];
    }
    
    /**
     * 
     * Retrieves the maximum z value in this PointList.  Will Take longer on the first call after a change has been made.
     * @return the largest z value in this PointList.
     */
    public float getZMax() {
        if (stats == null)
            makeStats();
        return stats[7];
    }
    
    /**
     * 
     * 
     * Retrieves the average z value in this PointList.  Will Take longer on the first call after a change has been made.
     * @return The average of the z values in this PointList.
     */
    public float getZAvg() {
        if (stats == null)
            makeStats();
        return stats[8];
    }
    
    /**
     * 
     * 
     * Retrieves the minimum w value in this PointList.  Will Take longer on the first call after a change has been made.
     * @return the smallest w value in this PointList.
     */
    public float getWMin() {
        if (stats == null)
            makeStats();
        return stats[9];
    }
    
    /**
     * 
     * Retrieves the maximum w value in this PointList.  Will Take longer on the first call after a change has been made.
     * @return the largest w value in this PointList.
     */
    public float getWMax() {
        if (stats == null)
            makeStats();
        return stats[10];
    }
    
    /**
     * 
     * Retrieves the average w value in this PointList.  Will Take longer on the first call after a change has been made.
     * @return The average of the w values in this PointList.
     */
    public float getWAvg() {
        if (stats == null)
            makeStats();
        return stats[11];
    }
    
    /**
     * 
     * Retreives the largest of any of the coordinates in this PointList. (i.e. max of getXMax,getYMax,getZMax,and getWMax)
     * @return the largest value in this PointList.
     */
    public float getMaxMax() {
        if (stats == null)
            makeStats();
        return Math.max(Math.max(stats[10],stats[7]),Math.max(stats[4],stats[1]));
    }
    
    /**
     * 
     * Retreives the smallest of any of the coordinates in this PointList. (i.e. min of getXMin,getYMin,getZMin,and getWMin)
     * @return the smallest value in this PointList.
     */
    public float getMinMin() {
        if (stats == null)
            makeStats();
        return Math.min(Math.min(stats[9],stats[6]),Math.min(stats[3],stats[0]));
    }
    
    /**
     * Generates a list of the points in this PointList
     * @return a string with all the points on their own line
     */
    public String toString() {
        StringBuffer sb = new StringBuffer("Point List with Elements:\n");
        
        for ( float[] f : list) {
            sb.append(f[0]);
            sb.append(", ");
            sb.append(f[1]);
            sb.append(", ");
            sb.append(f[2]);
            sb.append(", ");
            sb.append(f[3]);
            sb.append('\n');
        }
        
        return sb.toString();
    }

    /**
     * Determines if these point lists are element-by-element equal.
     * @param obj a PointList to compare to this (if not, guarenteed to return false)
     * @return true if the elements are all shared and the PointList 
     */
    public boolean equals(Object obj) {
        boolean b = false;
        if (obj instanceof PointList) {
            PointList plo = (PointList)obj;
            b = this.list.equals(plo.list);
        }
        return b;
    }

    /**
     * Generates a hash code using the backing objects' technique.
     * @return a value fulfilling the hashCode contract.
     */
    public int hashCode() {
        return list.hashCode();
    }
    
    /**
     * Copies this point list element by element.
     * @return a deep copy of this PointList
     */
    public PointList clone() {
        PointList ret = new PointList();
        for(float[] f : this.list)
            ret.addPoint(f);
        
        if (this.stats != null)
            ret.stats = this.stats.clone();
        else
            ret.stats = null;
        
        return ret; 
    }
    
    private void makeStats() {
        //stats = {xmin,xmax,xavg,ymin,ymax,yavg,zmin,zmax,zavg,wmin,wmax,wavg}
        stats = new float[12];
        
        stats[0] = stats[3] = stats[6] = stats[9] = Float.MAX_VALUE;
        stats[1] = stats[4] = stats[7] = stats[10] = Float.MIN_VALUE;
        stats[2] = stats[5] = stats[8] = stats[11] = 0.0f;
        for (float[] point : this.list) {
            //x-coord
            if (point[0] < stats[0])
                stats[0] = point[0];
            if (point[0] > stats[1])
                stats[1] = point[0];
            stats[2]+=point[0];
            //y-coord
            if (point[1] < stats[3])
                stats[3] = point[1];
            if (point[1] > stats[4])
                stats[4] = point[1];
            stats[5]+=point[1];
            //z-coord
            if (point[2] < stats[6])
                stats[6] = point[2];
            if (point[2] > stats[7])
                stats[7] = point[2];
            stats[8]+=point[2];
            //w-coord
            if (point[3] < stats[9])
                stats[9] = point[3];
            if (point[3] > stats[10])
                stats[10] = point[3];
            stats[11]+=point[3];
        }
        
        //normalize average
        stats[2]/=list.size();
        stats[5]/=list.size();
        stats[8]/=list.size();
        stats[11]/=list.size();
    }

    
    private class ElemIter implements Iterator<Float>{
        Iterator<float[]> listIter;
        int i;
        
        ElemIter(int index) {
            listIter = list.iterator();
            this.i = index;
        }
        
        public void remove() {
            listIter.remove();
        }
        
        public Float next() {
            return listIter.next()[i];
        }
        
        public boolean hasNext() {
            return listIter.hasNext();
        }
    }
    
    
    
}
